program Hole
!!$=head1 NAME
!!$
!!$Hole - Punch ellipsoidal hole in 2-D data
!!$
!!$=head1 SYNOPSIS
!!$
!!$Hole < data.H > hole.H
!!$
!!$=head1 DESCRIPTION
!!$
!!$Hole's dimensions and orientation are currently fixed
!!$
!!$=head1 PARAMETERS 
!!$
!!$=over 3
!!$
!!$=item from history
!!$
!!$B<integer> I<n1, n2>
!!$
!!$=back
!!$
!!$=head1 SEE ALSO
!!$
!!$L<Make>
!!$
!!$=cut
  use sep

  implicit none
  integer                            :: n1, n2, i1, i2, punchit, wantmask
  real, dimension (:,:), allocatable :: pp
  real                               :: x,y,u,v

  call sep_init ()
  call from_history (n1, n2)
  call from_either( "punchit", punchit, 1)
  call from_either( "wantmask", wantmask, 0)
  call sep_close ()

  allocate (pp(n1,n2) )
  call sep_read (pp)

	if( punchit==1 ) then
	  do i1=1,n1 
 	    do i2=1,n2 
 	       x = (i1-1.)/n1 - .5
 	       y = (i2-1.)/n2 - .3
 	       u =  x+y
 	       v = (x-y) / 2.
 	       if( u**2 + v**2 < .15 ) then
						pp(i1,i2) = 0.
         else	
						if( wantmask==1) pp(i1,i2) = 1.
				 end if
 	     end do
 	 end do
	else
		if( wantmask==1) where( pp/=0.0) pp=1.0
	end if

  call sep_write (pp)
  deallocate (pp)
  call exit (0)
end program Hole

