module bill_utils
use createhelixmod
use helix
use nhelix
use regrid
use bound
use misinput
use cartesian
use nbound_bob_mod
use error_check
use solver_prc_mod
use solver_reg_mod
use nhelicon
use ilaplacf
use laplacf
use lint
use cgstep_mod

contains

! SINGLE SCALE SUBROUTINES

  subroutine bin2points      (binned,on,oo,od,mask,points,coord)
    real,    dimension(:) ::  binned,   oo,od,     points
    logical, dimension(:) ::                  mask
    real,    dimension(:,:) ::                            coord
    integer, dimension(:) ::         on
    integer                 :: i,j,ic(size(on))
    
    nd=count(mask)
    j=1
    do i=1,product(on)
       if (mask(i)==.true.) then
          points(j)=binned(i)
          call line2cart(on,i,ic)
          coord(j,:)=oo+((ic-1)*od)
          j=j+1
       end if
    end do
  end subroutine bin2points

  subroutine find_nod(n,o,   d,sc,nn,no,         nd)
    integer ::        n(:),    sc,nn(size(n)),i    
    real    ::          o(:),d(:),   no(size(n)),nd(size(n)),nmax(size(n)),nmin(size(n))
    nn=int(sc*n/n(1))
    nmin = o-d/2.
    nmax = o +d*(n-1.) +d/2.
    nd = (nmax-nmin)/nn
    no = (nmin*(nn-.5) + nmax/2.)/nn
  end subroutine find_nod

  subroutine createpatch(npatch,  n,patch)
    integer ::i,j,      npatch(:),n(:)
    integer, dimension(size(npatch)) :: cartpos,psize,cartpatch
    integer, pointer ::             patch(:)
    do i=1,size(npatch)
       if (mod(n(i),npatch(i))>0) then
          write(0,*)"n not cleanly divisible by # of patches",mod(npatch(i),n(i))
       end if
    end do
    psize = n/npatch
    do i=1,product(n)                        ! loop over fine scale
       call line2cart(n,i,cartpos)           ! switch to cartesian space
       do j=1,size(npatch)
          cartpatch(j) = int((cartpos(j)-1.)/psize(j))+1   ! determine original micropatch table
       end do
       call cart2line(npatch,cartpatch,patch(i))   ! switch back to linear space
    end do
  end subroutine createpatch

  subroutine switchpatch(patch,n,   cn,swpatch)
    integer :: i,j,k,finel,    n(:),cn(:)
    integer, pointer ::  patch(:),     swpatch(:)
    integer,dimension(size(n)) :: coarsec,finec
    do i=1,product(cn)                   ! loop over mesh in current scale
       call line2cart(cn,i,coarsec)      ! loop through coarser space
       do j=1,size(cn)                   ! now find corresponding micropatch (mapped in fine space)
          finec(j) = nint(real(coarsec(j))*real(n(j))/real(cn(j)))
       end do
       call cart2line(n,finec,finel) ! go back to linear space
       swpatch(i) = patch(finel) ! assigns a value from fine space aa%patch to coarser scale
    end do
  end subroutine switchpatch

  subroutine switchlags(n,   cn,aa)
    integer ::          n(:),cn(:)
    type(nfilter)::             aa
    type(filter) ::                bb
    call allocatehelix(bb,size(aa%hlx(1)%lag))
    bb%lag=aa%hlx(1)%lag
    call regridn(n,cn,bb)
    do i=1,size(aa%hlx)
       aa%hlx(i)%lag=bb%lag
    end do
  end subroutine switchlags

  function scalelags        (n,jump,nl,na,center) result(newlags)
    integer :: i, lag0, lag,        nl
    integer, dimension(:) :: n,jump,   na,center
    integer, dimension(:), pointer ::                        newlags
    integer, dimension(size(n)) :: ii

    write(0,*)"scalelags jump",jump
    allocate(newlags(nl))
    call cart2line(  na, center, lag0)
    do i = 1, nl
       call line2cart(na, i+lag0, ii)
       ii = ii - center
       ii = ii * jump
       call cart2line( n,  ii+1, lag)
       newlags( i) = lag - 1
    end do
  end function scalelags

  subroutine scalelagsr     (n,jump,nl,na,center,newlags,midways)
    integer :: i, lag0, lag,        nl
    integer, dimension(:) :: n,jump,   na,center
    integer, dimension(:),  pointer ::           newlags
    real,    dimension(:,:),pointer ::                   midways
    integer, dimension(size(n)) :: ii
    real,    dimension(size(n)) :: jj

    call cart2line(  na, center, lag0)
    do i = 1, nl
       call line2cart(na, ih+lag0, ii)
       ii = ii - center
       jj = real(ii) * jump
       ii = int(jj)
       midways(i,:) = jj - ii
       call cart2line( n,  ii+1, lag)
       newlags( i) = lag - 1
    end do
  end subroutine scalelagsr

  subroutine interlace_data(       data,ildata,nn,fn)
    real, dimension(:), pointer :: data,ildata
    integer, dimension(:) ::                   nn,fn
    integer, dimension(:), allocatable ::            ii, jj, ilace
    integer               ::  i, j

write(0,*)"ilace start"
call rmms(ildata,"idata")
call rmms(data,"data ")
    allocate(ii(size(nn)),jj(size(nn)),ilace(size(nn)))
    do i=1,size(nn)
       ilace(i)=nint(real(fn(i))/real(nn(i)))
    end do
write(0,*)"ilace",ilace
    do i=1,product(nn)
       call line2cart(nn,i,ii)
       jj=ii+((ii-1)*(ilace-1))
       call cart2line(nn,jj,j)
       ildata(j)=data(i)
    end do
    deallocate(ii,jj,ilace)
  end subroutine interlace_data
    
  subroutine interlace_mask   (nn,fn,ilmask)
    integer, dimension(:) :: nn,fn
    logical, dimension(:) ::     ilmask
    integer, dimension(:), allocatable ::            ii, jj, ilace
    integer               ::  i, j

    allocate(ii(size(nn)),jj(size(nn)),ilace(size(nn)))
    ilace=fn/nn
    do i=1,product(nn)
       call line2cart(nn,i,ii)
       jj=ii*ilace
       call cart2line(nn,jj,j)
       ilmask(j)=.true.
    end do
    deallocate(ii,jj,ilace)
  end subroutine interlace_mask

  subroutine interlace_pef(pef,nn,fn)
    type(nfilter) ::       pef
    integer,dimension(:)::     nn,fn
    integer,dimension(:),allocatable :: oldpatch,newpatch,ii, jj, ilace
    integer               ::  i, j, k

    allocate(oldpatch(product(nn)),newpatch(product(fn)),ii(size(nn)),jj(size(nn)),ilace(size(nn)))
    oldpatch=pef%pch
    do i=1,size(nn)
       ilace(i)=nint(real(fn(i))/real(nn(i)))
    end do
    do j=1,product(fn)
       call line2cart(fn,j,jj)
       do k=1,size(fn)
          ii(k)=nint(real(jj(k))/real(ilace(k)))
       end do
       call cart2line(nn,ii,i)
       if (i>size(oldpatch)) then
          write(0,*)" I died at j=",j,jj,ilace,ii
       end if
       newpatch(j)=oldpatch(i)
    end do
    deallocate(ii,jj,ilace,pef%pch)
    allocate(pef%pch(product(fn)))
    pef%pch=newpatch
    deallocate(newpatch,oldpatch)

    call nbound_bob(fn,pef)

  end subroutine interlace_pef

end module bill_utils
