module igrad
! N-D gradient with adjoint,  r= grad( p)
  use adj_mod
  implicit none
  integer, dimension(:), allocatable, private :: n 
  integer, private                            :: dim
  contains

  subroutine igrad_init ( n_in )
    integer, dimension (:)            :: n_in
    dim=size(n_in)
    allocate(n(dim))
    n = n_in
  end subroutine igrad_init

  function igrad_lop ( adj, add, p, r) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: p,r 
    call adjnull (adj,add,p,r )

    if (dim==1) call igrad1_lop2(adj,p,r )
    if (dim==2) call igrad2_lop2(adj,p,r )
    if (dim==3) call igrad3_lop2(adj,p,r )
    if (dim==4) call igrad4_lop2(adj,p,r )
    if (dim==5) call igrad5_lop2(adj,p,r )
    if (dim==6) call igrad6_lop2(adj,p,r )
    stat=0
  end function igrad_lop

  subroutine igrad1_lop2(adj,p,r)
    logical,intent(in)          :: adj 
    real, dimension (n(1))      :: p 
    real, dimension (n(1),dim)  :: r 
    integer i
    do i= 1, n(1)-1  
       if ( adj) then
          p(i+1)  =		p(i+1)  +  r(i,1) 
          p(i  )  =                p(i  )  -  r(i,1) 
       else
          r(i,1)  =   r(i,1)  +  ( p(i+1) - p(i)) 
       end if
    end do
  end subroutine igrad1_lop2

  subroutine igrad2_lop2(adj,p,r)
    logical,intent(in)               :: adj 
    real, dimension (n(1),n(2))      :: p 
    real, dimension (n(1),n(2),dim)  :: r 
    integer i,j
    do i= 1, n(1)-1  
      do j= 1, n(2)-1  
         if ( adj) then
            p(i+1,j  )  =		 p(i+1,j  )  +  r(i,j,1) 
            p(i  ,j  )  =                p(i  ,j  )  -  r(i,j,1) 
            p(i  ,j+1)  =                p(i  ,j+1)  +  r(i,j,2) 
            p(i  ,j  )  =                p(i  ,j  )  -  r(i,j,2)
         else
            r(i,j,1)  =   r(i,j,1)  +  ( p(i+1,j) - p(i,j)) 
            r(i,j,2)  =   r(i,j,2)  +  ( p(i,j+1) - p(i,j))
         end if
      end do
   end do
 end subroutine igrad2_lop2

  subroutine igrad3_lop2(adj,p,r)
    logical,intent(in)                    :: adj 
    real, dimension (n(1), n(2), n(3))    :: p 
    real, dimension (n(1),n(2),n(3),dim)  :: r 
    integer i,j,k
    do i= 1, n(1)-1  
      do j= 1, n(2)-1  
         do k=1, n(3)-1
            if ( adj) then
               p(i+1,j  ,k  )  =		p(i+1,j  ,k  )  +  r(i,j,k,1) 
               p(i  ,j  ,k  )  =                p(i  ,j  ,k  )  -  r(i,j,k,1) 
               p(i  ,j+1,k  )  =                p(i  ,j+1,k  )  +  r(i,j,k,2) 
               p(i  ,j  ,k  )  =                p(i  ,j  ,k  )  -  r(i,j,k,2)
               p(i  ,j  ,k+1)  =                p(i  ,j  ,k+1)  +  r(i,j,k,3)
               p(i  ,j  ,k  )  =                p(i  ,j  ,k  )  -  r(i,j,k,3)
            else
               r(i,j,k,1)  =   r(i,j,k,1)  +  ( p(i+1,j,k) - p(i,j,k)) 
               r(i,j,k,2)  =   r(i,j,k,2)  +  ( p(i,j+1,k) - p(i,j,k))
               r(i,j,k,3)  =   r(i,j,k,3)  +  ( p(i,j,k+1) - p(i,j,k))
            end if
         end do
      end do
    end do
  end subroutine igrad3_lop2

  subroutine igrad4_lop2(adj,p,r)
    logical,intent(in)                            :: adj 
    real, dimension (n(1), n(2), n(3), n(4))      :: p 
    real, dimension (n(1), n(2), n(3), n(4),dim)  :: r 
    integer i,j,k,l
    do i= 1, n(1)-1  
      do j= 1, n(2)-1  
         do k=1, n(3)-1
            do l=1, n(4)-1
               if ( adj) then
                  p(i+1,j  ,k  ,l  )  =		p(i+1,j  ,k  ,l  )  +  r(i,j,k,l,1) 
                  p(i  ,j  ,k  ,l  )  =         p(i  ,j  ,k  ,l  )  -  r(i,j,k,l,1) 
                  p(i  ,j+1,k  ,l  )  =         p(i  ,j+1,k  ,l  )  +  r(i,j,k,l,2) 
                  p(i  ,j  ,k  ,l  )  =         p(i  ,j  ,k  ,l  )  -  r(i,j,k,l,2)
                  p(i  ,j  ,k+1,l  )  =         p(i  ,j  ,k+1,l  )  +  r(i,j,k,l,3)
                  p(i  ,j  ,k  ,l  )  =         p(i  ,j  ,k  ,l  )  -  r(i,j,k,l,3)
                  p(i  ,j  ,k  ,l+1)  =         p(i  ,j  ,k  ,l+1)  +  r(i,j,k,l,4)
                  p(i  ,j  ,k  ,l  )  =         p(i  ,j  ,k  ,l  )  -  r(i,j,k,l,4)
               else
                  r(i,j,k,l,1)  =   r(i,j,k,l,1)  +  ( p(i+1,j,k,l) - p(i,j,k,l)) 
                  r(i,j,k,l,2)  =   r(i,j,k,l,2)  +  ( p(i,j+1,k,l) - p(i,j,k,l))
                  r(i,j,k,l,3)  =   r(i,j,k,l,3)  +  ( p(i,j,k+1,l) - p(i,j,k,l))
                  r(i,j,k,l,4)  =   r(i,j,k,l,4)  +  ( p(i,j,k,l+1) - p(i,j,k,l))
               end if
            end do
         end do
      end do
    end do
  end subroutine igrad4_lop2

  subroutine igrad5_lop2(adj,p,r)
    logical,intent(in)                                  :: adj 
    real, dimension (n(1), n(2), n(3), n(4), n(5))      :: p 
    real, dimension (n(1), n(2), n(3), n(4), n(5),dim)  :: r 
    integer i,j,k,l,m
    do i= 1, n(1)-1  
      do j= 1, n(2)-1  
         do k=1, n(3)-1
            do l=1, n(4)-1
               do m=1, n(5)-1
                  if ( adj) then
                     p(i+1,j  ,k  ,l  ,m  )  =		p(i+1,j  ,k  ,l  ,m  )  +  r(i,j,k,l,m,1) 
                     p(i  ,j  ,k  ,l  ,m  )  =          p(i  ,j  ,k  ,l  ,m  )  -  r(i,j,k,l,m,1) 
                     p(i  ,j+1,k  ,l  ,m  )  =          p(i  ,j+1,k  ,l  ,m  )  +  r(i,j,k,l,m,2) 
                     p(i  ,j  ,k  ,l  ,m  )  =          p(i  ,j  ,k  ,l  ,m  )  -  r(i,j,k,l,m,2)
                     p(i  ,j  ,k+1,l  ,m  )  =          p(i  ,j  ,k+1,l  ,m  )  +  r(i,j,k,l,m,3)
                     p(i  ,j  ,k  ,l  ,m  )  =          p(i  ,j  ,k  ,l  ,m  )  -  r(i,j,k,l,m,3)
                     p(i  ,j  ,k  ,l+1,m  )  =          p(i  ,j  ,k  ,l+1,m  )  +  r(i,j,k,l,m,4)
                     p(i  ,j  ,k  ,l  ,m  )  =          p(i  ,j  ,k  ,l  ,m  )  -  r(i,j,k,l,m,4)
                     p(i  ,j  ,k  ,l  ,m+1)  =          p(i  ,j  ,k  ,l  ,m+1)  +  r(i,j,k,l,m,5)
                     p(i  ,j  ,k  ,l  ,m  )  =          p(i  ,j  ,k  ,l  ,m  )  -  r(i,j,k,l,m,5)
                  else
                     r(i,j,k,l,m,1)  =   r(i,j,k,l,m,1)  +  ( p(i+1,j,k,l,m) - p(i,j,k,l,m)) 
                     r(i,j,k,l,m,2)  =   r(i,j,k,l,m,2)  +  ( p(i,j+1,k,l,m) - p(i,j,k,l,m))
                     r(i,j,k,l,m,3)  =   r(i,j,k,l,m,3)  +  ( p(i,j,k+1,l,m) - p(i,j,k,l,m))
                     r(i,j,k,l,m,4)  =   r(i,j,k,l,m,4)  +  ( p(i,j,k,l+1,m) - p(i,j,k,l,m))
                     r(i,j,k,l,m,5)  =   r(i,j,k,l,m,5)  +  ( p(i,j,k,l,m+1) - p(i,j,k,l,m))
                  end if
               end do
            end do
         end do
      end do
   end do
 end subroutine igrad5_lop2

  subroutine igrad6_lop2(adj,p,r)
    logical,intent(in)                                        :: adj 
    real, dimension (n(1), n(2), n(3), n(4), n(5), n(6))      :: p 
    real, dimension (n(1), n(2), n(3), n(4), n(5), n(6),dim)  :: r 
    integer i,j,k,l,m,o
    do i= 1, n(1)-1  
      do j= 1, n(2)-1  
         do k=1, n(3)-1
            do l=1, n(4)-1
               do m=1, n(5)-1
                  do o=1, n(6)-1
                     if ( adj) then
                        p(i+1,j  ,k  ,l  ,m  ,o  )  =    p(i+1,j  ,k  ,l  ,m  ,o  )  +  r(i,j,k,l,m,o,1) 
                        p(i  ,j  ,k  ,l  ,m  ,o  )  =    p(i  ,j  ,k  ,l  ,m  ,o  )  -  r(i,j,k,l,m,o,1) 
                        p(i  ,j+1,k  ,l  ,m  ,o  )  =    p(i  ,j+1,k  ,l  ,m  ,o  )  +  r(i,j,k,l,m,o,2) 
                        p(i  ,j  ,k  ,l  ,m  ,o  )  =    p(i  ,j  ,k  ,l  ,m  ,o  )  -  r(i,j,k,l,m,o,2)
                        p(i  ,j  ,k+1,l  ,m  ,o  )  =    p(i  ,j  ,k+1,l  ,m  ,o  )  +  r(i,j,k,l,m,o,3)
                        p(i  ,j  ,k  ,l  ,m  ,o  )  =    p(i  ,j  ,k  ,l  ,m  ,o  )  -  r(i,j,k,l,m,o,3)
                        p(i  ,j  ,k  ,l+1,m  ,o  )  =    p(i  ,j  ,k  ,l+1,m  ,o  )  +  r(i,j,k,l,m,o,4)
                        p(i  ,j  ,k  ,l  ,m  ,o  )  =    p(i  ,j  ,k  ,l  ,m  ,o  )  -  r(i,j,k,l,m,o,4)
                        p(i  ,j  ,k  ,l  ,m+1,o  )  =    p(i  ,j  ,k  ,l  ,m+1,o  )  +  r(i,j,k,l,m,o,5)
                        p(i  ,j  ,k  ,l  ,m  ,o  )  =    p(i  ,j  ,k  ,l  ,m  ,o  )  -  r(i,j,k,l,m,o,5)
                        p(i  ,j  ,k  ,l  ,m  ,o+1)  =    p(i  ,j  ,k  ,l  ,m  ,o+1)  +  r(i,j,k,l,m,o,6)
                        p(i  ,j  ,k  ,l  ,m  ,o  )  =    p(i  ,j  ,k  ,l  ,m  ,o  )  -  r(i,j,k,l,m,o,6)
                     else
                        r(i,j,k,l,m,o,1)  =   r(i,j,k,l,m,o,1)  +  ( p(i+1,j,k,l,m,o) - p(i,j,k,l,m,o)) 
                        r(i,j,k,l,m,o,2)  =   r(i,j,k,l,m,o,2)  +  ( p(i,j+1,k,l,m,o) - p(i,j,k,l,m,o))
                        r(i,j,k,l,m,o,3)  =   r(i,j,k,l,m,o,3)  +  ( p(i,j,k+1,l,m,o) - p(i,j,k,l,m,o))
                        r(i,j,k,l,m,o,4)  =   r(i,j,k,l,m,o,4)  +  ( p(i,j,k,l+1,m,o) - p(i,j,k,l,m,o))
                        r(i,j,k,l,m,o,5)  =   r(i,j,k,l,m,o,5)  +  ( p(i,j,k,l,m+1,o) - p(i,j,k,l,m,o))
                        r(i,j,k,l,m,o,6)  =   r(i,j,k,l,m,o,6)  +  ( p(i,j,k,l,m,o+1) - p(i,j,k,l,m,o))
                     end if
                  end do
               end do
            end do
         end do
      end do
   end do
 end subroutine igrad6_lop2

  subroutine igrad_close()
    deallocate(n)
  end subroutine igrad_close

end module igrad
