module igrad1df
! Laplacian operator in N-D
  use adj_mod
  use igrad1
  implicit none

  logical, parameter,     private               :: T = .true. 
  logical, parameter,     private               :: F = .false. 
  integer, dimension (:), allocatable, private  :: n
  integer,                             private  :: nlag,n123,n1,n2,axis
  real, dimension (:), allocatable              :: temp
  real, dimension (:,:,:),allocatable             :: x2,y2


  contains
  subroutine igrad1df_init ( n1_in,n2_in,nlag_in,axis_in )
    integer            ::    n1_in,n2_in,nlag_in,axis_in
    n1 = n1_in
    n2 = n2_in
    nlag = nlag_in
    n123 = n1*n2
    axis = axis_in
    allocate (x2(nlag,n1,n2),y2(nlag,n1,n2),temp(nlag*n123))
  end subroutine

  function igrad1df_lop ( adj, add, x, y) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: x,y 
    call adjnull (adj,add,x,y )
    call igrad1df_lop2(adj,x,y )
    stat=0
  end function 

  subroutine igrad1df_lop2(adj,x,y)
    logical,intent(in) :: adj
    real, dimension (:)  :: x,y 
    integer stat1, ilag, i

    if ( adj) then
       y2 = reshape(y,(/nlag,n1,n2/))
       do ilag=1,nlag
          if (axis==1) then
             do i=1,n2
                stat1=igrad1_lop (  adj,  F, x2(ilag,:,i), y2(ilag,:,i))
             end do
          else
             do i=1,n1
                stat1=igrad1_lop (  adj,  F, x2(ilag,i,:), y2(ilag,i,:))
             end do
          end if
       end do
       temp = reshape(x2,(/nlag*n123/))
       x = x + temp
    else
       x2 = reshape(x,(/nlag,n1,n2/))
       do ilag=1,nlag
          if (axis==1) then
             do i=1,n2
                stat1 = igrad1_lop (  adj,  F, x2(ilag,:,i), y2(ilag,:,i))
             end do
          else
             do i=1,n1
                stat1 = igrad1_lop (  adj,  F, x2(ilag,:,i), y2(ilag,:,i))
             end do
          end if
       end do
       temp = reshape(y2,(/nlag*n123/))
       y = y + temp
    end if
  end subroutine igrad1df_lop2
  subroutine igrad1df_close()
    deallocate(x2,y2,temp)
  end subroutine 
end module 

