module ilaplacf
! Laplacian operator on 3D filter
  use adj_mod
  use ilaplacian
  implicit none

  logical, parameter,     private               :: T = .true. 
  logical, parameter,     private               :: F = .false. 
  integer, dimension (:), allocatable, private  :: m
  integer,                             private  :: nlag,m123 
  real, dimension (:), allocatable              :: temp
  real, dimension (:,:),allocatable             :: x2,y2

  contains

  subroutine ilaplacf_init ( m_in,   nlag_in,eps,na )
    integer            ::    m_in(:),nlag_in,    na
    real               ::                    eps
!write(0,*)"ilaplacf_init 1"
    allocate (m(size(m_in)))
!write(0,*)"ilaplacf_init 2"
    m = m_in
!write(0,*)"ilaplacf_init 3"
    nlag = nlag_in
!write(0,*)"ilaplacf_init 4",m,eps,na
    call ilaplac_init(m,eps,na)
!write(0,*)"ilaplacf_init 5"
    m123=product(m)
!write(0,*)"ilaplacf_init 6"
    allocate(x2(nlag,m123),y2(nlag,m123),temp(nlag*m123))
!write(0,*)"ilaplacf_init 7"
  end subroutine ilaplacf_init

  function ilaplacf_lop ( adj,add,x,y) result(stat)
    integer            ::                      stat 
    logical,intent(in) ::  adj,add
    real,dimension(:)  ::          x,y 
    call adjnull (adj,add,x,y )
    call ilaplacf_lop2(adj,x,y )
    stat=0
  end function ilaplacf_lop

  subroutine ilaplacf_lop2 (adj,x,y)
    logical,intent(in)   :: adj
    real, dimension (:)  ::     x,y 
    integer :: stat1, ilag

    if ( adj) then
       y2 = reshape(y,(/nlag,m123/))
       do ilag=1,nlag
          stat1 = ilaplac_lop (  adj,  F, x2(ilag,:), y2(ilag,:))
       end do
       temp = reshape(x2,(/nlag*m123/))
       x = x + temp
    else
       x2 = reshape(x,(/nlag,m123/))
       do ilag=1,nlag
          stat1 = ilaplac_lop (  adj,  F, x2(ilag,:), y2(ilag,:))
       end do
       temp = reshape(y2,(/nlag*m123/))
       y = y + temp
    end if
  end subroutine ilaplacf_lop2
  subroutine ilaplacf_close()
    deallocate(x2,y2,temp,m)
    call ilaplac_close()
  end subroutine ilaplacf_close
end module ilaplacf

