module ilaplacian
! Inverse laplacian in 2&3D
  use adj_mod
  use lapfac
  use polydiv
  use helix
  implicit none

  logical,   parameter, private   :: T = .true. 
  logical,   parameter, private   :: F = .false. 
  real ,private,  dimension (:), allocatable  :: tmp 
  type(filter)                    :: aa
  contains

  subroutine ilaplac_init ( n,eps,na )
    integer             ::  n(:), na
    real                ::    eps
!write(0,*)"ilaplac_init 1",n
    if (size(n)==2) then
!write(0,*)"ilaplac_init 2",eps,n(1),na,allocated(aa%lag)
       aa=lapfac2(eps,n(1),na)
!write(0,*)"ilaplac_init 3"
    elseif (size(n)==3) then
       aa=lapfac3(eps,n,na)
    else
       write(0,*)"Inverse Laplacian only supports 2&3D at this time"
    end if

!write(0,*)"ilaplac_init 4"
    call polydiv_init(product(n),aa)
!write(0,*)"ilaplac_init 5"
    allocate(tmp(product(n)))
!write(0,*)"ilaplac_init 6"
  end subroutine ilaplac_init

  function ilaplac_lop ( adj, add, x, y) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: x,y 
    call adjnull (adj,add,x,y )
    call ilaplac_lop2(adj,add,x,y )
    stat=0
  end function ilaplac_lop

  subroutine ilaplac_lop2 (adj,add,x,y)
    logical,intent(in) :: adj,add 
    real, dimension(:) ::         x,y
    integer stat1, stat2
    if ( adj) then
      stat1 = polydiv_lop (  T,  F, tmp, y)
      stat2 = polydiv_lop (  F,add, tmp, x)
    else
      stat1 = polydiv_lop (  T,  F, tmp, x)
      stat2 = polydiv_lop (  F,add, tmp, y)
    end if
  end subroutine ilaplac_lop2

  subroutine ilaplac_close()
write(0,*)"ilaplac close 1"
    deallocate( tmp )
write(0,*)"ilaplac close 2"
    call deallocatehelix(aa)
write(0,*)"ilaplac close 3"
    call polydiv_close
write(0,*)"ilaplac close 4"
  end subroutine ilaplac_close

end module ilaplacian
