module lapfac  
! Factor 2-D Laplacian.
  use wilson
  use error_check
  implicit none
  contains
  function lapfac2( eps, n1, na)   result (aa)
    type( filter)                       :: aa, lap
    real,                   intent( in) :: eps
    integer,                intent( in) :: n1, na
    integer                             :: i
    real                                :: a0, lap0
!write(0,*)"lapfac2 begin"

    call allocatehelix( lap, 2)         ! laplacian filter
!write(0,*)"lapfac2 2"
    lap0 = 4. + eps                     ! zero lag coeff.
!write(0,*)"lapfac2 3"
    lap%lag = (/ 1, n1 /)! lag(1)= 1; lag(2)=n1  # one side only
!write(0,*)"lapfac2 4"
    lap%flt = -1.! flt(1)=-1; flt(2)=-1
!write(0,*)"lapfac2 5"
    call allocatehelix( aa, 2*na)       ! laplacian derivative
!write(0,*)"lapfac2 6"
    aa%flt = 0.
!write(0,*)"lapfac2 7"
! probably done already in allocation.
    do i = 1, na  
      aa%lag( i   ) =      i! early lags (first row)
      aa%lag( na+i) = n1 + i - na! late lags (second row)
    end do 
!write(0,*)"lapfac2 8"
    call wilson_init( 10 * n1 )
!write(0,*)"lapfac2 9",20, lap0, a0
!call spefcheck(aa)
!call spefcheck(lap)
    call wilson_factor( 20, lap0, lap, a0, aa,.false.)
!write(0,*)"lapfac2 10"
    call wilson_close()
!write(0,*)"lapfac2 before deallocate"
    call deallocatehelix( lap)
!write(0,*)"lapfac2 end"
  end function lapfac2

  function lapfac3( eps,n,   na) result (aa)
    type(filter) ::                      aa, lap
    integer      ::     n(:),na, i
    real         :: eps,         a0, lap0

    call allocatehelix( lap, 3)
    lap0 = 6. + eps
    lap%lag = (/1, 1+n(1)*(n(2)-1), n(1)*n(2)/)
    call allocatehelix( aa, 6*na)
    aa%flt=0.
    do i=1,na
       aa%lag(i)=i
       aa%lag(5*na+i)=n(1)*(n(2)+1) - na + i
    end do
    do i=1,2*na
       aa%lag(i+na)=n(1) - na + i
       aa%lag(i+3*na)=n(1)*n(2) - na + i
    end do
    call wilson_init( 10*n(1)*n(2))
call spefcheck(aa)
call spefcheck(lap)
    call wilson_factor( 20, lap0, lap, a0, aa)
    call wilson_close()
!write(0,*)"lapfac2 before deallocate"
    call deallocatehelix(lap)
!write(0,*)"lapfac2 end"
  end function lapfac3

end module 
