module laplacf
! Laplacian operator in N-D
  use adj_mod
  use laplacian
  implicit none

  logical, parameter,     private               :: T = .true. 
  logical, parameter,     private               :: F = .false. 
  integer, dimension (:), allocatable, private  :: n
  integer,                             private  :: nlag,n123 
  real, dimension (:), allocatable              :: temp
  real, dimension (:,:),allocatable             :: x2,y2


  contains
  subroutine laplacf_init ( n_in,nlag_in )
    integer            :: n_in(:), nlag_in
    allocate (n(size(n_in)))
    n = n_in
    nlag = nlag_in
    call laplac_init (n)
    n123 = product(n)
    allocate (x2(nlag,n123),y2(nlag,n123),temp(nlag*n123))
  end subroutine

  function laplacf_lop ( adj, add, x, y) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: x,y 
    call adjnull (adj,add,x,y )
    call laplacf_lop2(adj,x,y )
    stat=0
  end function 

  subroutine laplacf_lop2(adj,x,y)
    logical,intent(in) :: adj
    real, dimension (:)  :: x,y 
    integer stat1, ilag

    if ( adj) then
       y2 = reshape(y,(/nlag,n123/))
       do ilag=1,nlag
          stat1 = laplac_lop (  adj,  F, x2(ilag,:), y2(ilag,:))
       end do
       temp = reshape(x2,(/nlag*n123/))
       x = x + temp
    else
       x2 = reshape(x,(/nlag,n123/))
       do ilag=1,nlag
          stat1 = laplac_lop (  adj,  F, x2(ilag,:), y2(ilag,:))
       end do
       temp = reshape(y2,(/nlag*n123/))
       y = y + temp
    end if
  end subroutine laplacf_lop2
  subroutine laplacf_close()
    deallocate(x2,y2,temp,n)
  end subroutine 
end module 

