module laplacian
! Laplacian operator in n-D
  use adj_mod
  use igrad
  implicit none
  logical,   parameter, private   :: T = .true. 
  logical,   parameter, private   :: F = .false. 
  real ,private,  dimension (:), allocatable  :: tmp 
  contains

  subroutine laplac_init ( n )
    integer, dimension(:)         :: n
    if (.not. allocated(tmp)) then
      allocate(tmp ( product(n)*size(n))) 
    end if
    call igrad_init (n)
  end subroutine laplac_init

  function laplac_lop ( adj, add, x, y) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: x,y 
    call adjnull (adj,add,x,y )
    call laplac_lop2(adj,x,y )
    stat=0
  end function laplac_lop

  subroutine laplac_lop2(adj,x,y)
    logical,intent(in) :: adj 
    real, dimension (:)  :: x 
    real, dimension (:)  :: y 
    integer stat1, stat2
    if ( adj) then
      stat1 = igrad_lop (  F,  F, y, tmp)   ! tmp = grad y
      stat2 = igrad_lop (  T,  T, x, tmp)   ! x = x + grad' tmp
    else
      stat1 = igrad_lop (  F,  F, x, tmp)   ! tmp = grad x
      stat2 = igrad_lop (  T,  T, y, tmp)   ! y = y + grad' tmp
    end if
  end subroutine laplac_lop2

  subroutine laplac_close()
    deallocate( tmp )
  end subroutine laplac_close

end module laplacian
