module nearf
! Nearest neighbor finder for NSPEFs
  use adj_mod
  use near2
  implicit none

  logical, parameter,     private               :: T = .true. 
  logical, parameter,     private               :: F = .false.
  logical, dimension (:,:),  pointer,  private  :: m
  integer, dimension (:,:,:),pointer,  private  :: map
  integer,                             private  :: nlag,n123,n1,n2
  real, dimension (:), allocatable              :: temp
  real, dimension (:,:),allocatable             :: x2,y2


  contains
  subroutine nearf_init ( m_in,map_in,n1_in,n2_in,nlag_in )
    integer                           :: nlag_in, n1_in, n2_in
    integer, dimension(:,:,:),pointer :: map_in
    logical, dimension(:,:),  pointer :: m_in
    n1=n1_in
    n2=n2_in
    nlag=nlag_in
    m => m_in
    map => map_in
    allocate (x2(nlag,n1*n2),y2(nlag,n1*n2),temp(nlag*n1*n2))
    call near2_init(m,map,n1,n2)
  end subroutine

  function nearf_lop ( adj, add, x, y) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: x,y 
    call adjnull (adj,add,x,y )
    call nearf_lop2(adj,x,y )
    stat=0
  end function 

  subroutine nearf_lop2(adj,x,y)
    logical, intent(in)  :: adj
    real, dimension (:)  :: x,y 
    integer stat1, ilag

    if ( adj) then
       y2 = reshape(y,(/nlag,n1*n2/))
       do ilag=1,nlag
          stat1 = near2_lop (  adj,  F, x2(ilag,:), y2(ilag,:))
       end do
       temp = reshape(x2,(/nlag*n1*n2/))
       x = x + temp
    else
       x2 = reshape(x,(/nlag,n1*n2/))
       do ilag=1,nlag
          stat1 = near2_lop (  adj,  F, x2(ilag,:), y2(ilag,:))
       end do
       temp = reshape(y2,(/nlag*n123/))
       y = y + temp
    end if
  end subroutine nearf_lop2
  subroutine nearf_close()
    deallocate(x2,y2,temp)
  end subroutine 
end module 

