module npef_find_nearer
! Find multi-scale prediction-error filter (helix magic)
  use bill_utils
  use nhconest
  use cgstep_mod
  use solver_prc_mod
  use solver_reg_mod
  use regnearf
  use dottest
  use error_check
  use nbound_bob_mod
  use nearf
  implicit none
contains

  subroutine npef_find1(data,  mask,n,o,   d,a,   center,   gap,   npatch,   niter,eps,prec,mod_res,dat_res,aa,epsl,na,vert,map,step)
    real,pointer    :: data(:),mask(:)
    real            ::                o(:),d(:),                                   eps,     mod_res(:),dat_res(:),   epsl
    integer         ::              n(:),    a(:),center(:),gap(:),npatch(:),niter,    prec,                na,bsize,nh,vert, step,i, j,r,nlag,i1,i2,l
    type(nfilter)   ::                                                                              aa
    integer, pointer:: patch(:)
    real,   pointer :: pefcoeffs(:)
    integer,pointer :: map(:,:,:)
    real, dimension(2):: dot1,dot2
    logical, pointer:: lmask(:),lmask2(:,:)

! __________________________________________________________
! ___________Find scale coordinates, size of arrays_________
! __________________________________________________________
    nh=size(aa%hlx(1)%flt)
    allocate(patch(product(n)))
write(0,*) "patch, npatch, n",npatch,n
    call createpatch(npatch,n,patch)
    aa%pch=patch
    allocate(pefcoeffs(nh*maxval(patch)))
!
!___________Calculate bounds
!
    allocate(lmask(size(mask)),lmask2(n(1),n(2)))
    lmask=.false.
    where (mask==0) lmask=.true. 
    call nbound_bob(n,aa,lmask)

write(0,*)"count/size aa%mis",count(aa%mis),size(aa%mis)
    lmask2=reshape(aa%mis,(/n(1),n(2)/))
!    lmask2=reshape(aa%mis,(/n(1),n(2)/))

write(0,*)"lmask2 details",count(lmask2),size(lmask2,1),size(lmask2,2)
!
! Precompute nearest neighbour map
!
!    allocate(map(n(1),n(2),2))
    do i=1,n(1)
       do j=1,n(2)
          if (lmask2(i,j)) then
             r=0
             do while(r<n(1))
                r=r+1
                if ((i+r+step>0) .and. (i+r+step<n(1))) then
                   if (.not. lmask2(i+r,j)) then
                      map(i,j,1)=i+r+step
                      map(i,j,2)=j
!                      write(0,*)"i,j,i+rr",i,j,i+r,map(i,j,1)
                      exit
                   end if
                end if
                if ((i-r-step>0) .and. (i-r<n(1))) then
                   if (.not. lmask2(i-r,j)) then
                      map(i,j,1)=i-r-step
                      map(i,j,2)=j
!                      write(0,*)"i,j,i-r",i,j,i+r,map(i,j,2)
                      exit
                   end if
                end if
                if (vert==0) then
                   if ((j+r>0) .and. (j+r+step<n(2))) then
                      if (.not. lmask2(i,j+r)) then
                         map(i,j,1)=i
                         map(i,j,2)=j+r+step
                         exit
                      end if
                   end if
                   if ((j-r-step>0) .and. (j-r<n(2))) then
                      if (.not. lmask2(i,j-r)) then
                         map(i,j,1)=i
                         map(i,j,2)=j-r-step
                         exit
                      end if
                   end if
                   if ((i+r>0) .and. (i+r+step<n(1)) .and. (j+r>0) .and. (j+r+step<n(2))) then
                      if (.not. lmask2(i+r,j+r)) then
                         map(i,j,1)=i+r+step
                         map(i,j,2)=j+r+step
                         exit
                      end if
                   end if
                   if ((i-r-step>0) .and. (i-r<n(1)) .and. (j-r-step>0) .and. (j-r-step<n(2))) then
                      if (.not. lmask2(i-r,j-r)) then
                         map(i,j,1)=i-r-step
                         map(i,j,2)=j-r-step
                         exit
                      end if
                   end if
                   if ((i+r>0) .and. (i+r+step<n(1)) .and. (j-r-step>0) .and. (j-r<n(2))) then
                      if (.not. lmask2(i+r,j-r)) then
                         map(i,j,1)=i+r+step
                         map(i,j,2)=j-r-step
                         exit
                      end if
                   end if
                   if ((i-r-step>0) .and. (i-r<n(1)) .and. (j+r>0) .and. (j+r+step<n(2))) then
                      if (.not. lmask2(i-r,j+r)) then
                         map(i,j,1)=i-r-step
                         map(i,j,2)=j+r+step
                         exit
                      end if
                   end if
                end if
             end do
          end if
       end do
    end do
write(0,*)"map computed",minval(map),maxval(map),size(map),count(map>0)
! __________________________________________________________
! ___________Pass precomputed stuff to operator init________
! __________________________________________________________

!call rmms(databank,"dbank")
    call nhconest_init(data,aa,nh)
!    call regnearf_init(lmask2,map,n(1),n(2),nh)
!    call nearf_init(lmask2,map,n(1),n(2),nh)
! __________________________________________________________
! ___________Call solver for either preconditioned or_______
! ___________just regularized case._________________________
! __________________________________________________________

    if (prec==1) then
!call rmms(databank,"dbank")
       call ilaplacf_init(npatch,nh,epsl,na)
!      call dot_test(nmhconest_new_lop,size(pefcoeffs),size(databank),dot1,dot2)
!      call dot_test(ilaplacf_lop,size(pefcoeffs),size(pefcoeffs),dot1,dot2)
       call solver_prc(m=pefcoeffs,                               &
                       d=-data,                               &
                      Fop=nhconest_lop,                      &
                      Sop=ilaplacf_lop,                           &
                      stepper=cgstep,                             &
                      nSop=(product(npatch)*size(aa%hlx(1)%flt)), &
                      niter=niter,                                &
                      eps=eps,                                    &
                      resd=dat_res,                               &
                      resm=mod_res,                               &
                      verb=.true.)
       call ilaplacf_close()
    else
       call laplacf_init(npatch,nh)
!      call dot_test(nhconest_lop,size(pefcoeffs),size(data),dot1,dot2)
!      call dot_test(regnearf_lop,size(pefcoeffs),size(pefcoeffs)*2,dot1,dot2)
!      call dot_test(nearf_lop,size(pefcoeffs),size(pefcoeffs),dot1,dot2)
!      call dot_test(laplacf_lop,size(pefcoeffs),size(pefcoeffs),dot1,dot2)
write(0,*)"eps",eps
      call solver_reg(m=pefcoeffs,                                &
                       d=-data,                                     &
                      Fop=nhconest_lop,                             &
                      Aop=laplacf_lop,                              &
!                      Aop=regnearf_lop,                              &
!                      Aop=nearf_lop,                                &
                      stepper=cgstep,                               &
                      nAop=(product(npatch)*size(aa%hlx(1)%flt)), &
                      niter=niter,                                  &
                      eps=eps,                                      &
!                      resd=dat_res,                                 &
!                      resm=mod_res,                                 &
                      verb=.true.)
!      call regnearf_close()
!      call nearf_close()
      call laplacf_close()
    end if
    call nhconest_close()
    call cgstep_close()
    
! __________________________________________________________
! ___________Re-initialize PEF for original data____________
! __________________________________________________________
write(0,*)"replace1",maxval(aa%pch),size(pefcoeffs)
do i=1,maxval(aa%pch)
   i1=mod(i,n(1))+1
   i2=int(floor(real(i)/real(n(1))))+1
   if (map(i1,i2,1)==0) then
      aa%hlx(i)%flt=pefcoeffs(1+(i-1)*nh:i*nh)
   else
      l=(map(i1,i2,2)-1)*n(1)+map(i1,i2,1)
      if (l>size(pefcoeffs)) write(0,*)"ohcrap",l
      aa%hlx(i)%flt=pefcoeffs(1+(l-1)*nh:l*nh)
      write(0,*)"aa%hlx(i%flt)",i,minval(aa%hlx(i)%flt),maxval(aa%hlx(i)%flt),size(aa%hlx(i)%flt)
   end if
end do
aa%mis=.false.
write(0,*)"replace10"

!    call set_pef(aa,patchbank,lagbank,boundbank,pefcoeffs,mark(2)-1,lagmark(2)-1)
! __________________________________________________________
! ___________Deallocate bookkeeping devices for scales______
! __________________________________________________________

    deallocate(pefcoeffs)
  end subroutine npef_find1

end module npef_find_nearer
