module npef_find_rnd
! Find multi-scale prediction-error filter (helix magic)
  use bill_utils
  use nhconest
  use cgstep_mod
  use solver_prc_mod
  use solver_reg_mod
  use ilaplacf
  use laplacf
  use dottest
  use error_check
  use nbound_bob_mod
  use random_mod
  implicit none
contains

  subroutine npef_find1(data,  mask,n,o,   d,a,   center,   gap,   npatch,   niter,eps,prec,mod_res,dat_res,aa,epsl,na,r0)
    real,pointer    :: data(:),mask(:)
    real            ::                o(:),d(:),                                   eps,     mod_res(:),dat_res(:),   epsl,r0(:)
    integer         ::              n(:),    a(:),center(:),gap(:),npatch(:),niter,    prec,                na,bsize,nh, i
    type(nfilter)   ::                                                                              aa
    integer, pointer:: patch(:)
    real,   pointer :: pefcoeffs(:),mask_rnd_resid(:)
    real, dimension(2):: dot1,dot2
    logical, pointer:: lmask(:)

! __________________________________________________________
! ___________Find scale coordinates, size of arrays_________
! __________________________________________________________
    nh=size(aa%hlx(1)%flt)
    allocate(patch(product(n)))
write(0,*) "patch, npatch, n",npatch,n
    call createpatch(npatch,n,patch)
    aa%pch=patch
    allocate(pefcoeffs(nh*maxval(patch)))
!
!
!  Initialize random model residual
    allocate(mask_rnd_resid(size(pefcoeffs)))
    mask_rnd_resid = r0

!
!___________Calculate bounds
!
    allocate(lmask(size(mask)))
    lmask=.false.
    where (mask==0) lmask=.true. 
    call nbound_bob(n,aa,lmask)

write(0,*)"count/size aa%mis",count(aa%mis),size(aa%mis)
! __________________________________________________________
! ___________Pass precomputed stuff to operator init________
! __________________________________________________________

!call rmms(databank,"dbank")
    call nhconest_init(data,aa,nh)
    
! __________________________________________________________
! ___________Call solver for either preconditioned or_______
! ___________just regularized case._________________________
! __________________________________________________________

    if (prec==1) then
!call rmms(databank,"dbank")
       call ilaplacf_init(npatch,nh,epsl,na)
!      call dot_test(nmhconest_new_lop,size(pefcoeffs),size(databank),dot1,dot2)
!      call dot_test(ilaplacf_lop,size(pefcoeffs),size(pefcoeffs),dot1,dot2)
       call solver_prc(m=pefcoeffs,                               &
                       d=-data,                               &
                      Fop=nhconest_lop,                      &
                      Sop=ilaplacf_lop,                           &
                      stepper=cgstep,                             &
                      nSop=(product(npatch)*size(aa%hlx(1)%flt)), &
                      niter=niter,                                &
                      eps=eps,                                    &
                      resd=dat_res,                               &
                      resm=mod_res,                               &
                      verb=.true.)
       call ilaplacf_close()
    else
       call laplacf_init(npatch,nh)
!      call dot_test(nmhconest_new_lop,size(pefcoeffs),size(databank),dot1,dot2)
!      call dot_test(laplacf_lop,size(pefcoeffs),size(pefcoeffs),dot1,dot2)
       call solver_reg(m=pefcoeffs,                               &
                       d=-data,                                   &
                      Fop=nhconest_lop,                           &
                      Aop=laplacf_lop,                            &
                      stepper=cgstep,                             &
                      nAop=(product(npatch)*size(aa%hlx(1)%flt)), &
                      niter=niter,                                &
                      eps=eps,                                    &
                      rm0=mask_rnd_resid,                         &
                      resd=dat_res,                               &
                      resm=mod_res,                               &
                      verb=.true.)
       call laplacf_close()
    end if
    call nhconest_close()
    call cgstep_close()
    
! __________________________________________________________
! ___________Re-initialize PEF for original data____________
! __________________________________________________________
do i=1,maxval(aa%pch)
   aa%hlx(i)%flt=pefcoeffs(1+(i-1)*nh:i*nh)
end do

!    call set_pef(aa,patchbank,lagbank,boundbank,pefcoeffs,mark(2)-1,lagmark(2)-1)
! __________________________________________________________
! ___________Deallocate bookkeeping devices for scales______
! __________________________________________________________

    deallocate(pefcoeffs)
  end subroutine npef_find1

end module npef_find_rnd
