module regnearf
! array of laplacian filter regularization and nearest-(masked)neighbour differencing
  use adj_mod
  use nearf
  use laplacf
  use chain_mod
  implicit none

  integer, dimension(2)                     :: n
  integer, private                          :: n1, n2, nlag
  logical, dimension(:,:), pointer, private :: m
  integer,dimension(:,:,:), pointer, private :: map
  real,   dimension(:), allocatable, private :: y_1, y_2

  contains
  subroutine regnearf_init ( m_in,map_in,n1_in,n2_in,nlag_in)
    integer            ::                n1_in,n2_in,nlag_in
    logical, dimension(:,:),pointer::m_in
    integer, dimension(:,:,:),pointer ::  map_in
    m => m_in
    map => map_in
    n1 = n1_in
    n2 = n2_in
    nlag = nlag_in
    n=(/n1,n2/)
    call laplacf_init(n,nlag)
    call nearf_init(m,map,n1,n2,nlag)
    allocate(y_1(n1*n2),y_2(n1*n2))
    y_1=0.; y_2=0.
write(0,*)"regnearf_init",n,nlag
  end subroutine regnearf_init

  function regnearf_lop ( adj, add, x, y) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: x,y 
    call adjnull (adj,add,x,y )
    call regnearf_lop2(adj,x,y )
    stat=0
  end function 

  subroutine regnearf_lop2(adj,x,y)
    logical, intent(in)  :: adj
    real, dimension (:)  :: x,y 
!    write(0,*)"Badj/mmsx/mmsy",adj,minval(x),maxval(x),size(x),minval(y),maxval(y),size(y)
    y_1=y(1:product(n))
    y_2=y(1+product(n):2*product(n))
    call array(laplacf_lop,nearf_lop,adj,.true.,x,y_1,y_2)
    y(1:product(n))=y_1
    y(1+product(n):2*product(n))=y_2
!    write(0,*)"Aadj/mmsx/mmsy",adj,minval(x),maxval(x),size(x),minval(y),maxval(y),size(y)
  end subroutine regnearf_lop2

  subroutine regnearf_close()
    call nearf_close()
    call laplacf_close()
    deallocate(y_1,y_2)
  end subroutine regnearf_close
end module regnearf
