module rescale_bill
use coarsen_new
use coarsen_norm_new
use createhelixmod
use helix
use nhelix
use regrid
use bound
use misinput
use cartesian
use nbound_bob_mod
use error_check
use solver_prc_mod
use solver_reg_mod
use nhelicon
use ilaplacf
use laplacf
use lint
use cgstep_mod

contains

! SINGLE SCALE SUBROUTINES

  subroutine bin2points      (binned,on,oo,od,mask,points,coord)
    real,    dimension(:) ::  binned,   oo,od,     points
    logical, dimension(:) ::                  mask
    real,    dimension(:,:) ::                            coord
    integer, dimension(:) ::         on
    integer                 :: i,j,ic(size(on))
    
    nd=count(mask)
    j=1
    do i=1,product(on)
       if (mask(i)==.true.) then
          points(j)=binned(i)
          call line2cart(on,i,ic)
          coord(j,:)=oo+((ic-1)*od)
          j=j+1
       end if
    end do
  end subroutine bin2points

  subroutine find_nod(n,o,   d,sc,nn,no,         nd)
    integer ::        n(:),    sc,nn(size(n)),i    
    real    ::          o(:),d(:),   no(size(n)),nd(size(n)),nmax(size(n)),nmin(size(n))
    nn=int(sc*n/n(1))
    nmin = o-d/2.
    nmax = o +d*(n-1.) +d/2.
    nd = (nmax-nmin)/nn
    no = (nmin*(nn-.5) + nmax/2.)/nn
  end subroutine find_nod

  function scale_mask  (mask,n, o,   d,nn,no,  nd) result(newmask)
    integer :: stat,         n(:),     nn(:)
    logical,pointer ::                                    newmask(:)
    real    ::          mask(:),o(:),d(:),no(:),nd(:)
    real,pointer:: tmask(:)
    allocate(tmask(product(nn)),newmask(product(nn)))
    tempmask=0.
    call coarsen_init(n,o,d,nn,no,nd)
call imms(n,"n    ")
call imms(nn,"nn   ")
call rmms(o,"o    ")
call rmms(no,"no   ")
call rmms(d,"d    ")
call rmms(nd,"nd   ")
    stat=coarsen_lop(.true.,.false.,tmask(1:product(nn)),mask)
    where (abs(tmask(1:product(nn)))>0)
       newmask=.true.
    end where
write(0,*)"count newmask",count(newmask)
    call coarsen_close()
    deallocate(tmask)
  end function scale_mask

  function scale_data(data,   mask,n,   o,   d,   nn,no,   nd) result(newdata)
    integer :: stat,               n(:),          nn(:)
    real,pointer ::   data(:),mask(:),                                newdata(:)
    real    ::                          o(:),d(:),   no(:),nd(:)
    allocate(newdata(product(nn)))
call rmms(data,"data ")
call rmms(mask,"mask ")
call imms(n,"n    ")
call imms(nn,"nn   ")
    call coarsen_norm_init(n,o,d,nn,no,nd,mask)
    stat=coarsen_norm_lop(.true.,.false.,newdata,data)
    call coarsen_norm_close()
call rmms(newdata,"ndata")
  end function scale_data

  function scale_data_pef(pef,data,   mask,n,o,d,nn,   no,   nd,   eps,fiter) result(newdata)
    integer :: stat,                       n(:), nn(:),                fiter
    real,pointer ::           data(:),mask(:),                                       newdata(:)
    real    ::                               o(:),d(:),no(:),nd(:),eps
    type(nfilter)::       pef
    logical,dimension(size(mask)):: lmask
    real,dimension(:,:),pointer  :: coord
    real,dimension(:),allocatable:: points
    integer,dimension(:),pointer :: coarsepatch
    allocate(newdata(product(nn)),coarsepatch(product(nn)))
    lmask=.false.
    where (mask>0) lmask=.true.
    allocate(points(count(lmask)),coord(count(lmask),size(n)))
    call bin2points(data,n,o,d,lmask,points,coord)
    call switchpatch(pef%pch,n,nn,coarsepatch)
    deallocate(pef%pch);allocate(pef%pch(product(nn)))
    pef%pch=coarsepatch
    call switchlags(n,nn,pef)
    call nbound_bob(nn,pef)
    call lint_init(nn,no,nd,coord)
    call nhelicon_init(pef)
    call solver_reg(m=newdata,            &
                    d=points,             &
                    Fop=lint_lop,         &
                    Aop=nhelicon_lop,     &
                    stepper=cgstep,       &
                    nAop=(product(nn)),   &
                    niter=fiter,          &
                    eps=eps)
    call nhelicon_close()
    call lint_close()
    call cgstep_close()
    deallocate(points,coord)
  end function scale_data_pef

  subroutine createpatch(npatch,  n,patch)
    integer ::i,j,      npatch(:),n(:)
    integer, dimension(size(npatch)) :: cartpos,psize,cartpatch
    integer, pointer ::             patch(:)
    do i=1,size(npatch)
       if (mod(n(i),npatch(i))>0) then
          write(0,*)"n not cleanly divisible by # of patches",mod(npatch(i),n(i))
       end if
    end do
    psize = n/npatch
    do i=1,product(n)                        ! loop over fine scale
       call line2cart(n,i,cartpos)           ! switch to cartesian space
       do j=1,size(npatch)
          cartpatch(j) = int((cartpos(j)-1.)/psize(j))+1   ! determine original micropatch table
       end do
       call cart2line(npatch,cartpatch,patch(i))   ! switch back to linear space
    end do
  end subroutine createpatch

  subroutine switchpatch(patch,n,   cn,swpatch)
    integer :: i,j,k,finel,    n(:),cn(:)
    integer, pointer ::  patch(:),     swpatch(:)
    integer,dimension(size(n)) :: coarsec,finec
    do i=1,product(cn)                   ! loop over mesh in current scale
       call line2cart(cn,i,coarsec)      ! loop through coarser space
       do j=1,size(cn)                   ! now find corresponding micropatch (mapped in fine space)
          finec(j) = nint(real(coarsec(j))*real(n(j))/real(cn(j)))
       end do
       call cart2line(n,finec,finel) ! go back to linear space
       swpatch(i) = patch(finel) ! assigns a value from fine space aa%patch to coarser scale
    end do
  end subroutine switchpatch

  subroutine switchlags(n,   cn,aa)
    integer ::          n(:),cn(:)
    type(nfilter)::             aa
    type(filter) ::                bb
    call allocatehelix(bb,size(aa%hlx(1)%lag))
    bb%lag=aa%hlx(1)%lag
    call regridn(n,cn,bb)
    do i=1,size(aa%hlx)
       aa%hlx(i)%lag=bb%lag
    end do
  end subroutine switchlags

  function scalelags        (n,jump,nl,na,center) result(newlags)
    integer :: i, lag0, lag,        nl
    integer, dimension(:) :: n,jump,   na,center
    integer, dimension(:), pointer ::                        newlags
    integer, dimension(size(n)) :: ii

    write(0,*)"scalelags jump",jump
    allocate(newlags(nl))
    call cart2line(  na, center, lag0)
    do i = 1, nl
       call line2cart(na, i+lag0, ii)
       ii = ii - center
       ii = ii * jump
       call cart2line( n,  ii+1, lag)
       newlags( i) = lag - 1
    end do
  end function scalelags

  subroutine scalelagsr     (n,jump,nl,na,center,newlags,midways)
    integer :: i, lag0, lag,        nl
    integer, dimension(:) :: n,jump,   na,center
    integer, dimension(:),  pointer ::           newlags
    real,    dimension(:,:),pointer ::                   midways
    integer, dimension(size(n)) :: ii
    real,    dimension(size(n)) :: jj

    call cart2line(  na, center, lag0)
    do i = 1, nl
       call line2cart(na, ih+lag0, ii)
       ii = ii - center
       jj = real(ii) * jump
       ii = int(jj)
       midways(i,:) = jj - ii
       call cart2line( n,  ii+1, lag)
       newlags( i) = lag - 1
    end do
  end subroutine scalelagsr

  subroutine interlace_data(       data,ildata,nn,fn)
    real, dimension(:), pointer :: data,ildata
    integer, dimension(:) ::                   nn,fn
    integer, dimension(:), allocatable ::            ii, jj, ilace
    integer               ::  i, j

write(0,*)"ilace start"
call rmms(ildata,"idata")
call rmms(data,"data ")
    allocate(ii(size(nn)),jj(size(nn)),ilace(size(nn)))
    do i=1,size(nn)
       ilace(i)=nint(real(fn(i))/real(nn(i)))
    end do
write(0,*)"ilace",ilace
    do i=1,product(nn)
       call line2cart(nn,i,ii)
       jj=ii+((ii-1)*(ilace-1))
       call cart2line(nn,jj,j)
       ildata(j)=data(i)
    end do
    deallocate(ii,jj,ilace)
  end subroutine interlace_data
    
  subroutine interlace_mask   (nn,fn,ilmask)
    integer, dimension(:) :: nn,fn
    logical, dimension(:) ::     ilmask
    integer, dimension(:), allocatable ::            ii, jj, ilace
    integer               ::  i, j

    allocate(ii(size(nn)),jj(size(nn)),ilace(size(nn)))
    ilace=fn/nn
    do i=1,product(nn)
       call line2cart(nn,i,ii)
       jj=ii*ilace
       call cart2line(nn,jj,j)
       ilmask(j)=.true.
    end do
    deallocate(ii,jj,ilace)
  end subroutine interlace_mask

  subroutine interlace_pef(pef,nn,fn)
    type(nfilter) ::       pef
    integer,dimension(:)::     nn,fn
    integer,dimension(:),allocatable :: oldpatch,newpatch,ii, jj, ilace
    integer               ::  i, j, k

    allocate(oldpatch(product(nn)),newpatch(product(fn)),ii(size(nn)),jj(size(nn)),ilace(size(nn)))
    oldpatch=pef%pch
    do i=1,size(nn)
       ilace(i)=nint(real(fn(i))/real(nn(i)))
    end do
    do j=1,product(fn)
       call line2cart(fn,j,jj)
       do k=1,size(fn)
          ii(k)=nint(real(jj(k))/real(ilace(k)))
       end do
       call cart2line(nn,ii,i)
       if (i>size(oldpatch)) then
          write(0,*)" I died at j=",j,jj,ilace,ii
       end if
       newpatch(j)=oldpatch(i)
    end do
    deallocate(ii,jj,ilace,pef%pch)
    allocate(pef%pch(product(fn)))
    pef%pch=newpatch
    deallocate(newpatch,oldpatch)

    call nbound_bob(fn,pef)

  end subroutine interlace_pef

! MULTISCALE ROUTINES (loop over scale, call single scale routines)

  function find_nbank(n,   sc) result(nbank)
    integer ::        n(:),sc(:), i,  nbank(size(n),size(sc))    
    do i=1,size(sc)
       nbank(:,i)=int(sc(i)*n(:)/n(1))
    end do
  end function find_nbank

  subroutine find_nodbank(n,o,d,sc,nbank,obank,dbank)
    integer ::            n(:),sc(:), i,  nbank(size(n),size(sc))    
    real    ::              o(:),d(:),obank(size(n),size(sc)),dbank(size(n),size(sc)),nmax(size(n)),nmin(size(n))
    do i=1,size(sc)
       call find_nod(n,o,d,sc(i),nbank(:,i),obank(:,i),dbank(:,i))
    end do
  end subroutine find_nodbank
  
  function find_mark(nbank) result(mark)
    integer :: i,    nbank(:,:)
    integer ::              mark(size(nbank,2)+1)
    mark(1)=1
    do i=1,size(nbank,2)
       mark(i+1)=mark(i)+product(nbank(:,i))
    end do
  end function find_mark

  function find_lagmark(aa,sc) result(lagmark)
    type(filter) ::     aa
    integer      ::        sc(:)
    integer      ::                   lagmark(size(sc)+1) ! output
    lagmark(1)=1
    do i=1,size(sc)
       lagmark(i+1)=lagmark(i)+size(aa%lag)
    end do
  end function find_lagmark

  subroutine maskbank_find(mask,n,   o,   d,nbank,obank,     dbank,maskbank)
    integer :: stat,          n(:),       nbank(:,:),mark(size(nbank,2)+1)
    logical,pointer ::                                             maskbank(:)
    real    :: tempmask(:),mask(:),o(:),d(:),   obank(:,:),dbank(:,:)
    allocate(tempmask(product(n)))
    mark=find_mark(nbank)
    call imms2(nbank,"nbank")
    call imms(mark,"mark ")
    write(0,*)"mark",mark
    call lmms(maskbank,"mbank")
    do i=1,size(nbank,2)
       maskbank(mark(i):mark(i+1)-1)=scale_mask(mask,n,o,d,nbank(:,i),obank(:,i),dbank(:,i))
    end do
  end subroutine maskbank_find

  subroutine databank_find(data,   mask,n,   o,   d,   nbank,obank,     dbank,databank)
    integer :: stat,                  n(:),          nbank(:,:), mark(size(nbank,2)+1)
    real,pointer ::      data(:),mask(:),                                     databank(:)
    real    ::                             o(:),d(:),      obank(:,:),dbank(:,:)
    mark=find_mark(nbank)
write(0,*)"mark",mark
    do i=1,size(nbank,2)
       databank(mark(i):mark(i+1)-1)=scale_data(data,mask,n,o,d,nbank(:,i),obank(:,i),dbank(:,i))
write(0,*)"rescaling:",i," of",size(nbank,2)
    end do
call rmms(databank,"dbank")
  end subroutine databank_find

  subroutine databank_pef_find(pef,data,   mask,n,   o,   d,   nbank,obank,     dbank,eps,fiter,     databank)
    integer :: stat,i,                          n(:),          nbank(:,:), fiter,mark(size(nbank,2)+1)
    real,pointer ::                data(:),mask(:)
    real    ::                                       o(:),d(:),      obank(:,:),dbank(:,:),eps,databank(:) ! output
    type(nfilter)::            pef
    logical,dimension(size(mask))::lmask
    real,dimension(:,:),pointer:: coord
    real,dimension(:),allocatable:: points
    integer,dimension(:),allocatable::oldpatch

    mark=find_mark(nbank)
    allocate(oldpatch(size(pef%pch)))
    oldpatch=pef%pch
    do i=1,size(nbank,2)
       databank(mark(i):mark(i+1)-1)=scale_data_pef(pef,data,mask,n,o,d,nbank(:,i),obank(:,i),dbank(:,i),eps,fiter)
       if (allocated(pef%pch)) deallocate(pef%pch); allocate(pef%pch(size(oldpatch))); pef%pch=oldpatch
       call switchlags(nbank(:,i),n,pef)
    end do
    deallocate(oldpatch)
  end subroutine databank_pef_find

  subroutine boundbank_find(maskbank,nbank,     a,   center,   gap,boundbank)
    logical         ::    maskbank(:)
    logical,pointer ::                                             boundbank(:)
    integer         ::             nbank(:,:),a(:),center(:),gap(:),mark(size(nbank,2)+1)
    type(filter)    :: aa

    mark=find_mark(nbank)
    allocate(boundbank(mark(size(mark))-1))
    aa = createhelix(nbank(:,1), center, gap, a)
    boundbank=.false.
    do i=1,size(nbank,2)
       call regridn (nbank(:,1),nbank(:,i),aa)
       call boundn (nbank(:,i),nbank(:,i),a,aa)
       call find_mask (maskbank(mark(i):mark(i+1)-1),aa)
       boundbank(mark(i):mark(i+1)-1)=aa%mis
       deallocate(aa%mis)
       call regridn (nbank(:,i),nbank(:,1),aa)
    end do
    call deallocatehelix(aa)
  end subroutine boundbank_find

  subroutine boundbank_sean_find(n,mask,lagbank,   nlag,a,   center,   gap,boundbank)
    logical, intent(in) ::         mask(:)
    integer, intent(in) ::       n(:),  lagbank(:),nlag,a(:),center(:),gap(:)
    logical, intent(out)::                                                 boundbank(:)
    type(filter)        :: aa
    integer :: nj
write(0,*)"bound sean 0",nlag
    nj=size(lagbank)/nlag
write(0,*)"bound sean 1"
    aa = createhelix(n,center,gap,a)
write(0,*)"bound sean 2"
    boundbank=.false.
    if (nlag/=size(aa%lag)) write(0,*)"error in lag sizes:boundbank_sean_find"
write(0,*)"bound sean 3"
    do i=1,nj
       aa%lag=lagbank(1+(i-1)*nlag:i*nlag)
       call boundn(n,n,a,aa)
       call find_mask(mask,aa)
       write(0,*)"Scale:",i," Equations:",count(.not. aa%mis)," of ",size(aa%mis)
       boundbank(1+(i-1)*nlag:i*nlag)=aa%mis
       deallocate(aa%mis)
    end do

  end subroutine boundbank_sean_find


  subroutine micropatchbank_find (npatch,   nbank,patchbank)
    integer ::i,                  npatch(:),nbank(:,:),    mark(size(nbank,2)+1)
    integer, pointer ::                           patchbank(:),tpb(:),tpb2(:)
    mark=find_mark(nbank)
    allocate(tpb(mark(2)-1))
    call createpatch(npatch,nbank(:,1),tpb)
    patchbank(mark(1):mark(2)-1)=tpb

write(0,*)"woohoo mark",mark
call imms(patchbank,"pbank")

    do i=2,size(nbank,2)
       allocate(tpb2(mark(i+1)-mark(i)))
       call switchpatch(tpb,nbank(:,1),nbank(:,i),tpb2)
       patchbank(mark(i):mark(i+1)-1)=tpb2
       deallocate(tpb2)
    end do
  end subroutine micropatchbank_find
  
  subroutine lagbank_find (nbank,     a,   center,   gap,lagbank,   lagmark)
    integer,intent(in) ::  nbank(:,:),a(:),center(:),gap(:)                    ! inputs
    integer,intent(out)::                                lagbank(:),lagmark(:) ! outputs
    type(filter) :: aa
    write(0,*) "nbank",nbank
    aa = createhelix(nbank(:,1), center, gap, a)
    lagbank=0
    lagmark=find_lagmark(aa,nbank(1,:))
    call imms(lagmark,"lmark")
    call imms2(nbank,"nbank")
    call imms(a,"    a")
    call imms(center,"cente")
    call imms(gap,"  gap")
    call imms(lagbank,"lbank")
    call spefcheck(aa)
    do i=1,size(nbank,2) 
       call regridn(nbank(:,1),nbank(:,i),aa) 
       lagbank(lagmark(i):lagmark(i+1)-1)=aa%lag
       call regridn(nbank(:,i),nbank(:,1),aa)
    end do
write(0,*) "Through lagbank before deallocate"
    call deallocatehelix(aa)
  end subroutine lagbank_find

  subroutine lagbank_jump_find (jbank,     n,   a,   center,   gap,   nlags,lagbank)
    integer,intent(in)  ::      jbank(:,:),n(:),a(:),center(:),gap(:),nlags
    integer,intent(out) ::                                                  lagbank(:)
    integer             :: i

    do i=1,size(jbank,1)
       lagbank(1+(i-1)*nlags:i*nlags)=scalelags(n,jbank(i,:),nlags,a,center+gap)
    end do
  end subroutine lagbank_jump_find

  subroutine set_pef(aa,patchbank,   lagbank,boundbank,pefcoeffs,mark,lagmark)
    type(nfilter)  ::aa
    integer,pointer::   patchbank(:),lagbank(:)      
    real,   pointer::                                  pefcoeffs(:)
    integer        ::                                            mark,lagmark,i,nf,nh
    logical,pointer::                        boundbank(:)

    if (allocated(aa%mis)) then
       deallocate(aa%mis)
    end if
    allocate(aa%mis(mark))

    if(allocated(aa%pch)) then
       deallocate(aa%pch)
    end if
       allocate(aa%pch(mark))
    aa%mis=boundbank(1:mark)
    aa%pch=patchbank(1:mark)
    nh=size(aa%hlx)
    nf=size(aa%hlx(1)%flt)
    do i=1,nh
       aa%hlx(i)%lag = lagbank(1:lagmark)
       aa%hlx(i)%flt = pefcoeffs((1+(i-1)*nf):(1+(i)*nf))
    end do
  end subroutine set_pef

end module rescale_bill
